QBL = {
	version = "3.11 for Workgroups (3.11.6)",

	-- Colorized tag for status messages.
	_q = "|cffBE81F7[|r|cffFFF600Q?|r|cffBE81F7]|r",

	-- Used to label add-on messages over add-on chnnel.
	prefix = "[Q?]",

	-- Variable table to pass around flags between modules. Used sparingly.
	var = {},

	-- Utility modules, called throughout the program.
	util = {},

	-- Librariy modules. Only exported vars and functions use this table.
	lib = {},

	-- Libraries define their event handlers in this table.
	events = {},

	-- Always-on event listeners, registered at startup.
	startup = {},

	-- Service table for registering and unregistering event listeners.
	srv = {
		event = {},
		opt = {},
		on = {}
	},

	-- You may provide a 'hook' to call when starting or stopping 'srv'.
	-- See lib/Maps for a simple example.
	hook = {},

	-- Locale and table for holding internationalization strings.
	locale = GetLocale(),
	i18n = {enUS = {}},

	-- Default options. Initializes the QBL_cfg table.
	cfg_defaults = {
		-- Questing
		tracker = true,
		print_quests = false,
		auto_quests = 4,
		accept_trivial = false,
		do_dailies = 2,
		auto_reward = false,
		-- Friends
		follow_friend = true,
		accept_invite = true,
		promote_friend = false,
		invite_friend = true,
		bnet_friends = false,
		-- Trash
		auto_sell = true,
		guild_repair = false,
		drop_trash_loot = true,
		drop_trash_reward = true,
		keep_gray = false,
		keep_de = false,
		-- Loot
		fast_loot = 2,
		stacker = true,
		reagent = true,
		stack_gbank = false,
		loot_left = false,
		-- Interface
		mini_coord = true,
		world_coord = true,
		reverse_meta = false,
		confirm_de = false,
		sound = true,
		intro = true
	},

	-- Main Event and Configuration Menu frames.
	frames = {
		main = CreateFrame("Frame", "QBL_frame"),
		menu = CreateFrame("Frame", "QBL_menu", UIParent)
	}
}

RegisterAddonMessagePrefix(QBL.prefix)

--[[ Schema for QBL.srv, maintained by srv_mt.
	event = {
		TRADE_CLOSED = { "option1", "option3" },
		MAIL_CLOSED = { "option1", "option2" },
		PLAYERBANKSLOTS_CHANGED = { "option3" }
	},
	opt = {
		option1 = { "TRADE_CLOSED", "MAIL_CLOSED" },
		option2 = { "MAIL_CLOSED" },
		option3 = { "TRADE_CLOSED", "PLAYERBANKSLOTS_CHANGED" }
	},
	on = {
		option1 = false,
		option2 = false,
		option3 = false
	}
--]]

-- Transforms assignments to QBL.srv.option to the above schema.
local srv_mt = {
	__newindex = function(tbl, key, val)
		tbl.opt[key] = val
		tbl.on[key] = false
		for _, srv in ipairs(val) do
			if not tbl.event[srv] then tbl.event[srv] = {key}
			else table.insert(tbl.event[srv], key) end
		end
	end
}
setmetatable(QBL.srv, srv_mt)

-- Register and unregister events. Run hook functions needed by an option.
function QBL:enable_opt(option)
	if self.srv.opt[option] then
		for _, service in ipairs(self.srv.opt[option]) do
			local run = true
			if self.srv.event[service] then
				for _, opt in ipairs(self.srv.event[service]) do
					if opt ~= option and self.srv.on[opt] == true then
						run = false
						break
					end
				end
			end
			if run then self.frames.main:RegisterEvent(service) end
		end
	end
	if self.hook[option] then self.hook[option](true, self.frames.main) end
	self.srv.on[option] = true
end

function QBL:disable_opt(option)
	if self.srv.opt[option] then
		for _, service in ipairs(self.srv.opt[option]) do
			local run = true
			if self.srv.event[service] then
				for _, opt in ipairs(self.srv.event[service]) do
					if opt ~= option and self.srv.on[opt] == true then
						run = false
						break
					end
				end
			end
			if run then self.frames.main:UnregisterEvent(service) end
		end
	end
	if self.hook[option] then self.hook[option](false, self.frames.main) end
	self.srv.on[option] = false
end

-- Default to 'enUS' if localization is missing.
local i18n_mt = {__index = QBL.i18n.enUS}
setmetatable(QBL.i18n, i18n_mt)

-- Used to send text 'msg' to other Q? users in a group.
-- See lib\Groups for further information.
function QBL:send_msg(msg)
	local type
	if IsInRaid() then type = "RAID"
	elseif IsInGroup() then type = "PARTY"
	else return false end
	SendAddonMessage(self.prefix, msg, type)
	return true
end

-- Used to copy the cfg_defaults table when needed.
function QBL:copy_cfg()
	local copy = {}
	for k, v in pairs(self.cfg_defaults) do copy[k] = v end
	return copy
end

-- Print command that directs Q? messages to local chat or its own chat frame.
function QBL:print(...)
	local tab
	for i = 1, 10 do
		local name, _, _, _, _, _, shown, _, docked = GetChatWindowInfo(i)
		if (shown or docked) and name == "Q?" then
			tab = i
			break
		end
	end

	if tab then
		local msg = ""
		for i = 1, select("#", ...) do msg = msg .. " " .. select(i, ...) end
		_G["ChatFrame" .. tab]:AddMessage(msg)
	else print(self._q, ...) end
end
