local i18n = QBL.i18n[QBL.locale].lib.Quests

local fmt_money = QBL.util.Text.fmt_money
local fmt_notice = QBL.util.Text.fmt_notice
local is_auto_ok = QBL.util.Config.is_auto_ok
local wowhead_url = QBL.util.Links.wowhead_url
local is_inv_full = QBL.util.Trash.is_inv_full
local find_lowest_trash = QBL.util.Trash.find_lowest_trash

-- Frame used to register timer events for stacking.
local w_frame = getglobal("WorldFrame")

-- Returns the quest reward worth the most money.
local function select_best(choices)
	local best_choice, highest_val = 0, 0
	for choice = 1, choices do
		local item = GetQuestItemLink("choice", choice)
		local _, _, quantity = GetQuestItemInfo("choice", choice)
		if item then
			local price = select(11, GetItemInfo(item))
			if price * quantity > highest_val then
				best_choice = choice
				highest_val = price * quantity
			end
		end
	end
	return best_choice
end

-- Automatically accepts quests and rewards, if appropriate, otherwise
-- it highlights the best reward value.
local function handle_reward()
	local choices = GetNumQuestChoices()
	if QBL_cfg.auto_reward then
		if choices == 0 then
			GetQuestReward(QuestFrameRewardPanel.itemChoice)
		elseif choices == 1 and QuestFrameRewardPanel.itemChoice == nil
		then GetQuestReward(1) else
			local best_choice = select_best(choices)
			if best_choice > 0 then GetQuestReward(best_choice) end
		end
	elseif QBL_cfg.auto_quests == 3 or QBL_cfg.auto_quests == 4 then
		if choices == 0 then
			GetQuestReward(QuestFrameRewardPanel.itemChoice)
		elseif choices == 1 and QuestFrameRewardPanel.itemChoice == nil
		then GetQuestReward(1) else
			local best_choice = select_best(choices)
			if best_choice > 0 then
				_G["QuestInfoRewardsFrameQuestInfoItem"
					.. best_choice].Name:SetTextColor(0.92, 0.81, 0.25)
			end
		end
	end
end

-- Called via OnUpdate, it will delete 'count' many trash items, if needed.
local trash_count, trash_state = 0, 0
local function drop_trash(self)
	if trash_count > 0 and trash_state == 1
		and not CursorHasItem()
	then
		local link, bag, slot = find_lowest_trash()
		if bag and slot
		then
			QBL:print(fmt_notice(i18n.REPLACE_TRASH, "WARN"), link)
			PickupContainerItem(bag, slot)
		else trash_count = 0 end
	elseif trash_count > 0 and trash_state == 1
		and CursorHasItem()
	then trash_state = 2
	elseif trash_count > 0 and trash_state == 2 then
		DeleteCursorItem()
		trash_state = 3
	elseif trash_count > 0 and trash_state == 3 then
		trash_count = trash_count - 1
		trash_state = 1
	elseif trash_count == 0 then
		self:SetScript("OnUpdate", nil)
		trash_count = 0
		trash_state = 0
		handle_reward()
	end
end

-- Auto-accept rewards, otherwise highlight the most valuable reward.
-- Will print quest text and drop trash if nessesary.
function QBL.events:QUEST_COMPLETE(...)
	if QBL_cfg.print_quests then
		local npc = UnitName("questnpc")
		if npc then QBL:print(npc) end
		QBL:print(GetRewardText())
	end
	local choices = GetNumQuestChoices()

	if is_auto_ok() then
		if QBL_cfg.drop_trash_reward then
			local _, free = is_inv_full()
			local _, _, _, trash = find_lowest_trash()

			local count = 0
			if choices > 0 then
				count = GetNumQuestRewards() + 1
			else count = GetNumQuestRewards() end

			if count > 0 and count > free then
				if count > trash then
					QBL:print(i18n.TRASH_ALERT)
					trash_count = trash
				else trash_count = count - free end

				trash_state = 1
				w_frame:SetScript("OnUpdate", function(self) drop_trash(self) end)
			else handle_reward() end
		else handle_reward() end
	else
		if choices > 1 then
			local best_choice = select_best(choices)
			if best_choice > 0 then
				_G["QuestInfoRewardsFrameQuestInfoItem"
					.. best_choice].Name:SetTextColor(0.92, 0.81, 0.25)
			end
		end
	end
end

-- Print quest info and WoWhead link. Drop trivial quests auto-picked up.
-- Some quest offers do not have a GOSSIP_SHOW or QUEST_GREETING event.
-- Level or triviality information cannot be obtained from QUEST_DETAIL.
-- We drop trivial quests unless auto_ok while this event fires.
function QBL.events:QUEST_ACCEPTED(idx, ...)
	local _, level, _, _, _, _, _, id, _, _, _, _, task = GetQuestLogTitle(idx)
	local lvl_diff = level - UnitLevel("player")
	SelectQuestLogEntry(idx)

	-- 'task' indicates an area given quest, which by process of elimination is
	-- either a Bonus Objective or World Quest.
	local is_world = QuestUtils_IsQuestWorldQuest(id)
	local is_bonus = false
	if task and not is_world then is_bonus = true end

	-- need_quest is only true if a user has intentionally requested a quest.
	-- As such, Q? should never drop it, even if trivial.
	if (QBL_cfg.auto_quests == 2 or QBL_cfg.auto_quests == 4)
		and is_auto_ok()
		and not is_bonus and not is_world
		and not QBL.var.need_quest and not QBL_cfg.accept_trivial
		and not (lvl_diff >= -2 or -lvl_diff <= GetQuestGreenRange())
	then
		QBL:print(i18n.DROP_Q .. ":", wowhead_url(idx))
		QBL:print(i18n.DROP_HELP)
		SetAbandonQuest(idx)
		AbandonQuest()
	else
		-- Clear this, as quest has been accepted. SEE: util\Links
		QBL.var.need_quest = nil
		QBL:print("URL:", wowhead_url(idx))

		local money, xp = GetQuestLogRewardMoney(), GetQuestLogRewardXP()
		if money > 0 and xp > 0 then
			QBL:print(i18n.REWARD..":", fmt_money(money), "|cff00FF00("..xp.." XP)|r")
		elseif money > 0 and xp <= 0 then
			QBL:print(i18n.REWARD..":", fmt_money(money))
		elseif money <= 0 and xp > 0 then
			QBL:print(i18n.REWARD..":", "|cff00FF00("..xp.." XP)|r")
		end

		local curr_count = GetNumQuestLogRewardCurrencies()
		local curr_txt = ""
		if curr_count > 0 then
			for i = 1, curr_count do
				name, _, amount = GetQuestLogRewardCurrencyInfo(i, id)
				curr_txt = curr_txt .. "|cffFFFFCC" .. name
					.. "|r |cff00FF00(+" .. amount .. ")|r"
				if i < curr_count then
					curr_txt = curr_txt .. ", "
				end
			end
		end
		if curr_txt ~= "" then QBL:print(i18n.CURR_REWARD .. ":", curr_txt) end

		-- Bonus Objectives and World Quests do not provide correct faction info
		-- for some reason.
		if not is_bonus and not is_world then
			local fact_count = GetNumQuestLogRewardFactions()
			local fact_txt = ""
			if fact_count == 1 then
				local id, amount = GetQuestLogRewardFactionInfo(1);
				local name = GetFactionInfoByID(id);
				fact_txt = "|cffFFFFCC" .. name
					.. "|r |cff00FF00(+" .. amount / 100 .. ")|r"
			elseif fact_count > 1 then
				for i = 1, fact_count do
					local id, amount = GetQuestLogRewardFactionInfo(i)
					local name
					if GetFactionInfoByID(id) then name = GetFactionInfoByID(id)
					else name = "--" end
					fact_txt = fact_txt .. "|cffFFFFCC" .. name
						.. "|r |cff00FF00(+" .. amount / 100 .. ")|r"
					if i < fact_count then
						fact_txt = fact_txt .. ", "
					end
				end
			end
			if fact_txt ~= "" then QBL:print(i18n.FACT_REWARD .. ":", fact_txt) end
		end
	end
end

-- Returns true if quest 'title' is daily.
local function is_daily(title)
	local active, g_active = GetNumActiveQuests(), GetNumGossipActiveQuests()
	if active > 0 then
		for i = 1, active do
			local name = GetActiveTitle(i)

			if name == title then
				local _, daily = GetAvailableQuestInfo(i)
				if daily ~= 1 then daily = true else daily = false end

				return daily
			end
		end
	elseif g_active > 0 then
		for i = 1, g_active do
			local name
			if i == 1 then name = select(1, GetGossipActiveQuests())
			else name = select(((6 * (i - 1)) + 1), GetGossipActiveQuests()) end

			if name == title then
				local daily
				if i == 1 then daily = select(4, GetGossipActiveQuests())
				else daily = select(((6 * (i - 1)) + 4), GetGossipActiveQuests()) end
				if daily ~= 1 then daily = true else daily = false end

				return daily
			end
		end
	end
end

-- Detects options the UI has deemed trivial via tinting the 'icon'.
local function is_trivial(icon)
	local r, g, b = icon:GetVertexColor()
	return math.floor(r * 256) == 128
		and math.floor(g * 256) == 128
		and math.floor(b * 256) == 128
end

-- Clicks a 'button' of the correct 'type' and 'icon'.
-- The quest 'title' is used for verifying quest turn-ins.
-- What is considered correct is based upon Q? settings.
local function scan_btn(button, type, icon, title)
	-- We do not handle repetables, too dangerous to users.
	if type == "DailyActiveQuestIcon" then return
	elseif type == "ActiveQuestIcon"
	then
		if not is_trivial(icon) or QBL_cfg.accept_trivial then
			if (QBL_cfg.auto_quests == 3 or QBL_cfg.auto_quests == 4)
				and not is_daily(title)
			then button:Click()
			elseif (QBL_cfg.do_dailies == 3 or QBL_cfg.do_dailies == 4)
				and is_daily(title)
			then button:Click() end
		end
	elseif (QBL_cfg.do_dailies == 2 or QBL_cfg.do_dailies == 4)
		and type == "DailyQuestIcon"
	then
		if not is_trivial(icon) or QBL_cfg.accept_trivial then button:Click() end
	elseif (QBL_cfg.auto_quests == 2 or QBL_cfg.auto_quests == 4)
		and (type == "AvailableQuestIcon"
			or type == "AvailableLegendaryQuestIcon")
	then
		if not is_trivial(icon) or QBL_cfg.accept_trivial then button:Click() end
	elseif (QBL_cfg.auto_quests == 3 or QBL_cfg.auto_quests == 4)
		and type == "ActiveLegendaryQuestIcon"
	then
		if not is_trivial(icon) or QBL_cfg.accept_trivial then button:Click() end
	end
end

-- Strip color codes from strings.
local function strip(str)
	str = gsub(str, "|c%x%x%x%x%x%x%x%x", "")
	str = gsub(str, "|r", "")
	return str
end

function QBL.events:GOSSIP_SHOW(...)
	if QBL.var.need_quest then SelectGossipAvailableQuest(1)
	elseif is_auto_ok() then
		local button
		for i = 1, 32 do
			button = _G['GossipTitleButton' .. i]
			if button:IsVisible() then
				local icon = _G[button:GetName() .. "GossipIcon"]
				local type = string.sub(icon:GetTexture(), 23)
				local title = strip(button:GetText())
				scan_btn(button, type, icon, title)
			end
		end
	end
end

function QBL.events:QUEST_GREETING(...)
	if QBL.var.need_quest then SelectAvailableQuest(1)
	elseif is_auto_ok() then
		local button
		for i = 1, 32 do
			button = _G['QuestTitleButton' .. i]
			if button:IsVisible() then
				local icon = _G[button:GetName() .. "QuestIcon"]
				local type = string.sub(icon:GetTexture(), 23)
				local title = strip(button:GetText())
				scan_btn(button, type, icon, title)
			end
		end
	end
end

function QBL.events:QUEST_DETAIL(idx, ...)
	if QBL_cfg.print_quests then
		local npc = UnitName("questnpc")
		if npc then QBL:print(npc) end
		QBL:print(GetQuestText())
	end
	if (QBL_cfg.auto_quests == 2 or QBL_cfg.auto_quests == 4)
		and not QuestIsDaily() and not QuestIsWeekly()
		and is_auto_ok()
	then AcceptQuest()
	elseif (QBL_cfg.do_dailies == 2 or QBL_cfg.do_dailies == 4)
		and (QuestIsDaily() or QuestIsWeekly())
		and is_auto_ok()
	then AcceptQuest() end
end

function QBL.events:QUEST_PROGRESS(...)
	if QBL_cfg.print_quests then
		local npc = UnitName("questnpc")
		if npc then QBL:print(npc) end
		QBL:print(GetProgressText())
	end
	if (QBL_cfg.auto_quests == 3 or QBL_cfg.auto_quests == 4)
		and not QuestIsDaily() and not QuestIsWeekly()
		and is_auto_ok()
	then CompleteQuest()
	elseif (QBL_cfg.do_dailies == 3 or QBL_cfg.do_dailies == 4)
		and (QuestIsDaily() or QuestIsWeekly())
		and is_auto_ok()
	then CompleteQuest() end
end

-- Prints WowHead links when user meta-clicks Qust Tracker quests.
function QBL.events:SUPER_TRACKED_QUEST_CHANGED(idx, ...)
	if IsModifierKeyDown()
	then QBL:print("URL:", wowhead_url(GetQuestLogIndexByID(idx))) end
end

QBL.srv.auto_quests = {"GOSSIP_SHOW", "QUEST_GREETING", "QUEST_DETAIL",
	"QUEST_PROGRESS"}
QBL.srv.print_quests = {"QUEST_DETAIL", "QUEST_PROGRESS"}
QBL.srv.do_dailies = {"GOSSIP_SHOW", "QUEST_GREETING", "QUEST_DETAIL",
	"QUEST_PROGRESS"}

QBL.startup["QUEST_ACCEPTED"] = true
QBL.startup["QUEST_COMPLETE"] = true
QBL.startup["SUPER_TRACKED_QUEST_CHANGED"] = true
